#!/usr/bin/env python3
""" Mobosearch project: Json Web converter
Creates a website from json and templates

SPDX-FileCopyrightText: 2022 Massimo Marchiori <massimo@math.unipd.it>
SPDX-License-Identifier: GPL-3.0-or-later
"""
__author__ = "Massimo Marchiori"
__version__ = "1.15"

import os
import datetime
import shutil

import simplejson as json
import ijson
import re
import argparse
import gzip

# Default values (can be modified to set new defaults)
DIR_BASE = "."  # Base working directory
DIR_JSON = "google/ranked"  # Directory containing the json source(s)
DIR_TEMPLATES = "templates"  # Directory containing the templates
DIR_WEBSITE = "website"  # Directory for the generated website
MAGIC_BEGIN = "[[["  # Magic sequence activating JsonWeb command
MAGIC_END = "]]]"  # End sequence for JsonWeb command
ALLOWED_EXTENSIONS = [".html", ".htm"]  # allowed extensions for input pages
DISALLOWED_EXTENSIONS = ["~"]  # disallowed extensions for input pages
COPY_FILES = False  # if true, files that are not in (allowed+disallowed) are copied
FILE_COMPRESS = False  # if true, processed (allowed) pages are compressed with gzip

# JsonWeb command prefixes
COMMAND_CONTEXT = "context:"  # Case-insensitive
COMMAND_FILENAME = "filename:"  # Case-insensitive


def find_commands(disassembled, command, base, file, unique):
    """ Find JsonWeb commands """
    commands = [i for i, e in enumerate(disassembled) if e and e.strip().lower().startswith(command)]
    if unique and len(commands) > 1:
        print(f"Malformed template {os.path.join(base, file)}:")
        print(f'Too many commands "{command}" ({len(commands)})')
        exit(0)
    return commands


def get_command_parameter(disassembled, command, command_index):
    """ Find the parameters corresponding to a command """
    if len(command_index):
        parameter = disassembled[command_index[0]].strip()[len(command):].strip()
    else:
        parameter = None
    return parameter


def path_to_ijson(path):
    """ Translate a normal path into an ijson path """
    if path is None:
        return ""
    else:
        return path.replace("*", "item").replace("'", "").replace('"', "").replace("[", "").replace("]", "")


def extract(object, path):
    """ Extract information with the specific path """
    if isinstance(path, str) and path.strip() in ["", "."]:
        return object
    elif path is None:
        return ""
    steps = path.split(".")
    parts = [object]
    for step in steps:
        new_parts = []
        for part in parts:
            part_type = type(part)
            if step == "*":
                if part_type is list:
                    new_parts += part
                elif part_type is dict:
                    new_parts += part.items()
                else:
                    new_parts.append(part)
            else:
                if part_type is dict:
                    if step in part:
                        new_parts.append(part[step])
                    else:
                        if len(parts) == 1:
                            new_parts.append(None)
                elif part_type in [list, tuple]:
                    if step.isdigit():
                        if int(step) < len(part):
                            new_parts.append(part[int(step)])
                        else:
                            new_parts.append(None)
                    else:
                        new_parts.append(None)
                else:
                    new_parts.append(None)
        parts = new_parts
    if len(parts) == 1:
        return parts[0]
    else:
        return parts


def write_page(web_file, disassembled, values):
    """ Write the page with processed command values """
    for i, e in enumerate(disassembled):
        if i % 2:
            value = values[(i-1)//2]
            web_file.write(json.dumps(value)[1:-1] if type(value) is str else json.dumps(value))
        else:
            web_file.write(disassembled[i])


def process_file(base, file, extension):
    """ Process a template file """
    with open(os.path.join(base, file), "r", encoding="utf-8") as file_template:
        template = file_template.read()
        disassembled = re.split(re.escape(MAGIC_BEGIN) + "|" + re.escape(MAGIC_END), template)
        if len(disassembled) % 2 == 0:
            print(f"Malformed template {os.path.join(base, file)}:")
            print(f"Mismatch with separators {MAGIC_BEGIN} and {MAGIC_END}")
            exit(0)

        contexts = find_commands(disassembled, COMMAND_CONTEXT, base, file, True)
        context = get_command_parameter(disassembled, COMMAND_CONTEXT, contexts)
        if context:
            disassembled[contexts[0]] = None

        filename_commands = find_commands(disassembled, COMMAND_FILENAME, base, file, True)
        filename_param = get_command_parameter(disassembled, COMMAND_FILENAME, filename_commands)
        if filename_param:
            disassembled[filename_commands[0]] = None

    file_counter = 0
    for j in os.scandir(DIR_JSON):
        with open(j.path, "r", encoding='utf-8') as json_file:
            objects = ijson.items(json_file, path_to_ijson(context), multiple_values=True)
            for obj in objects:
                file_counter += 1
                values = []
                filename = extract(obj, filename_param)
                if not filename:
                    filename = file
                elif filename == "*":
                    filename = os.path.splitext(file)[0] + "_" + str(file_counter) + os.path.splitext(file)[1]
                odd_position = False
                for path in disassembled:
                    if odd_position:
                        odd_position = False
                        values.append(extract(obj, path))
                    else:
                        odd_position = True
                if not filename.endswith(extension):
                    filename += extension
                final_filename = os.path.normpath(
                    os.path.join(DIR_BASE,
                                 DIR_WEBSITE,
                                 os.path.relpath(base, os.path.join(DIR_BASE, DIR_TEMPLATES)),
                                 filename))
                if FILE_COMPRESS:
                    with gzip.open(final_filename + ".gz", "wt", encoding='utf-8') as web_file:
                        write_page(web_file, disassembled, values)
                else:
                    with open(final_filename, "w", encoding='utf-8') as web_file:
                        write_page(web_file, disassembled, values)


def file_endswith(file, extensions):
    """ Check if a filename ends with a certain extension """
    for suffix in extensions:
        if file.endswith(suffix):
            return suffix
    return False


if __name__ == '__main__':
    print(f"JsonWeb (v{__version__}) - the static website generator from JSON data")
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", "--input", default=DIR_JSON, help="the input json directory")
    parser.add_argument("-t", "--templates", default=DIR_TEMPLATES, help="the templates directory")
    parser.add_argument("-o", "--output", default=DIR_WEBSITE, help="the output web directory")
    parser.add_argument("-b", "--base", default=DIR_BASE, help="base directory for relative directory paths")
    parser.add_argument("-l", "--left", default=MAGIC_BEGIN,
                        help=f"left delimiter of JsonWeb command activator")
    parser.add_argument("-r", "--right", default=MAGIC_END,
                        help=f"right delimiter of JsonWeb command activator")
    parser.add_argument("-a", "--allowed", nargs='*', default=ALLOWED_EXTENSIONS,
                        help=f"allowed file extensions: files are processed")
    parser.add_argument("-d", "--disallowed", nargs='*', default=DISALLOWED_EXTENSIONS,
                        help=f"disallowed file extensions: files are ignored")
    parser.add_argument("-c", "--copy", action='store_true', default=COPY_FILES,
                        help="copy all other files (not in allowed+disallowed)")
    parser.add_argument("-g", "--gzip", action='store_true', default=FILE_COMPRESS,
                        help="compress allowed output pages with gzip (adds .gz suffix)")
    args = parser.parse_args()

    DIR_JSON = os.path.normpath(args.input)
    DIR_TEMPLATES = os.path.normpath(args.templates)
    DIR_WEBSITE = os.path.normpath(args.output)
    DIR_BASE = os.path.normpath(args.base)
    MAGIC_BEGIN = args.left
    MAGIC_END = args.right
    ALLOWED_EXTENSIONS = args.allowed
    DISALLOWED_EXTENSIONS = args.disallowed
    COPY_FILES = args.copy
    FILE_COMPRESS = args.gzip

    for extension in args.allowed:
        if not extension:
            print("ERROR: empty extension for allowed files: aborting")
            exit(0)
        if extension[0] != ".":
            print(f"WARNING: allowed extension '{extension}' doesn't start with a dot")

    print(f"** Starting site generation at {datetime.datetime.now()}")
    for base, dirs, files in os.walk(os.path.join(DIR_BASE, DIR_TEMPLATES)):
        for directory in dirs:
            os.makedirs(os.path.join(DIR_BASE, DIR_WEBSITE, directory), exist_ok=True)
        for file in files:
            if file_endswith(file, DISALLOWED_EXTENSIONS):
                continue
            extension = file_endswith(file, ALLOWED_EXTENSIONS)
            if extension:
                process_file(base, file, extension)
            else:
                if COPY_FILES:
                    shutil.copy(os.path.join(base, file),
                                os.path.normpath(
                                    os.path.join(DIR_BASE,
                                                 DIR_WEBSITE,
                                                 os.path.relpath(base, os.path.join(DIR_BASE, DIR_TEMPLATES)),
                                                 file)))

    print(f"** Completed site generation at {datetime.datetime.now()}")
